// application color theme

class ColorTheme{
	constructor(){
		this._theme_list = [
			'day',
			'night',
		];
		this._color_resource = {
			'day': {
			'cid_bg': '#f0f0f0',
			'cid_fg': '#333',
			'cid_light': '#ccc',
			'cid_input': '#aaa',
			'cid_edit': '#bbd',
			'cid_del': '#dbb',
			'cid_box': '#ddd',
			'cid_boxshadow': '5px 5px 10px 0px rgba(0,0,0,0.25)',
			'cid_link': '#88f',
			'cid_pnbg': '#ccc',
			'cid_pnfg': '#333',
			'cid_cnbg': '#55b',
			'cid_cnfg': '#eee',
			'cid_filter': 'brightness(100%)',
			'cid_logofill': '#333',
			'cid_logostroke': '#f0f0f0',
			'cid_twsfill': 'none',
			'cid_twsstroke': '#333',
			'cid_pststroke': '#333',

			'cid_colorings': ['bw', 'hue', 'line',],
			'cid_fills': {
				'bw': ['#333',],
				'line':['none',],
				'hue':['#c3d825', '#82ae46', '#d7003a', '#38a1db', '#ffdb4f', '#59b9c6', '#ee827c',
					 '#bc64a4', '#38b48b', '#ffec47', '#e95295', '#e6b422', '#ea5506', '#4d5aaf', ],
				},
			'cid_strokes': {
				'bw': ['#333',],
				'line':['#333',],
				'hue':['#fff'],
				},
			
			},
			'night':{
			'cid_bg': '#000',
			'cid_fg': '#888',
			'cid_light': '#333',
			'cid_input': '#666',
			'cid_edit': '#282848',
			'cid_del': '#482828',
			'cid_box': '#223',
			'cid_boxshadow': '0px 0px 10px 0px rgba(240,240,255,0.25) inset',
			'cid_link': '#447',
			'cid_pnbg': '#181818',
			'cid_pnfg': '#585858',
			'cid_cnbg': '#181840',
			'cid_cnfg': '#778',
			'cid_filter': 'brightness(50%)',
			'cid_logofill': '#666',
			'cid_logostroke': '#000',
			'cid_twsfill': 'none',
			'cid_twsstroke': '#888',
			'cid_pststroke': '#888',
			
			'cid_colorings': ['bw', 'hue', 'line',],
			'cid_fills': {
				'bw': ['#666',],
				'line':['none',],
				'hue':['#626c13', '#415723', '#6c001d', '#1c516e', '#806e28', '#2d5d63', '#77413e',
					 '#5e3252', '#1c5a46', '#807624', '#75294b', '#735a11', '#752b03', '#272d58', ],
				},
			'cid_strokes': {
				'bw': ['#666',],
				'line':['#888',],
				'hue':['#181818',],
				},
			
			},
		};
		var date = new Date();
		var hour = date.getHours();
		this._cuurentThemeid = ( hour >= 20 || hour <= 7 ) ? 1 : 0;
	}

	_findCssRule(exp){
		for (var i = 0; i < document.styleSheets.length; i++) {
			var rules = document.styleSheets[i].cssRules;
			for (var j = 0; j< rules.length; j++) {
				if( rules[j].selectorText == exp ){
					return rules[j];
				}
			}
		}
		return {};
	}

	toggleTheme(){
		this._cuurentThemeid = (this._cuurentThemeid + 1) % this._theme_list.length;
		this.applyTheme();
	}

	applyTheme(){
		var res = this._color_resource[ this._theme_list[this._cuurentThemeid] ];
		this._findCssRule('body').style['background-color'] = res['cid_bg'];
		this._findCssRule('body').style['color'] = res['cid_fg'];
		this._findCssRule('.material-icons.dispicon').style['color'] = res['cid_light'];
		this._findCssRule('.material-icons.editicon').style['color'] = res['cid_edit'];
		this._findCssRule('.material-icons.deletemode').style['color'] = res['cid_del'];
		this._findCssRule('.txtlink').style['color'] = res['cid_link'];
		this._findCssRule('.shadowbox').style['border-color'] = res['cid_box'];
		this._findCssRule('.shadowbox').style['box-shadow'] = res['cid_boxshadow'];
		this._findCssRule('.language').style['filter'] = res['cid_filter'];
		this._findCssRule('#patanimage').style['filter'] = res['cid_filter'];
		this._findCssRule('.pn').style['border-color'] = res['cid_pnbg'];
		this._findCssRule('.pn').style['color'] = res['cid_pnfg'];
		this._findCssRule('.cn').style['border-color'] = res['cid_cnbg'];
		this._findCssRule('.cn').style['background-color'] = res['cid_cnbg'];
		this._findCssRule('.cn').style['color'] = res['cid_cnfg'];
		this._findCssRule('#editsilhouettename').style['border-color'] = res['cid_light'];
		this._findCssRule('#editsilhouettename').style['background-color'] = res['cid_bg'];
		this._findCssRule('#editsilhouettename').style['color'] = res['cid_input'];
	}

	getCurrentThemeColor(cid){
		return this._color_resource[ this._theme_list[ this._cuurentThemeid ] ][cid];
	}

	getCurrentThemeColorings(){
		var res = this._color_resource[ this._theme_list[this._cuurentThemeid] ];
		return {
			'variation': res['cid_colorings'].length,
			'colorings': res['cid_colorings'],
			'fills': res['cid_fills'],
			'strokes': res['cid_strokes'],
		};
	}
}
