// patan edit canvas

class EditCanvas{
	constructor(patan, element){
		this._patan = patan;
		this._scope = new paper.PaperScope();
		this._scope.setup(element);

		//------ Paperobject Global Variables
		this._pieceNum = 0;
		this._coloring = 0;
		this._fills = [];
		this._strokes = [];
		this._currZoom = 1;
		this._zoomRatio = 1;
		this._strokeBaseWidth = 1.5;
		this._pathData = {};
		this._charLength = {};
		
		this._group = new this._scope.Group();
		this._edgeGroup = new this._scope.Group();
		this._pieceInfo = [];
		this._selectedInfo = [];
		this._initPositions = [];
		this._initAngles = [];
		this._initFlips = [];
		this._initCenter;
		
		this._UndoBuffer = function(positions, angles, flips){
			this.positions = positions;
			this.angles = angles;
			this.flips = flips; 
			return this;
		}
		this._undoSize = 250;
		this._undoBuffer = [];
		this._undoIndex = 0;
		this._undoStart = 0;
		this._undoLimit = 0;
		
		this._isAnimtaion = false;
		this._isSetUndoAfterAnimation = false;
		this._isExclusiveModifing = false;
		
		this._activePiecePid = -1;
		this._lastActivePiecePid = -1;

		//------ Internal logics
		this._animUndoDiv = 5;
		this._animRewinedDiv = 7;
		
		//------ Mouse operations
		this._animClosingDiv = 3;

		this._ratiodtheta;
		this._ratiodthetamin = 0.2;
		this._ratiodthetath = 0.5;

		this._flipDeltas = [];
		this._flipDeltaIndex;
		this._flipAccumLength = 40;
		this._flipAccumDelta;
		this._flipAccumAbsDelta;
		this._flipAbsDeltaMin = 50;
		this._flipAbsDeltaMax = 200;

		this._isDraged = false;
		this._isDragedMore = false;
		this._dragAccumAbsDelta;
		this._dragAccumAbsDth = 4;
		this._dragMoreAccumAbsDth = 20;
		this._rectStartPoint;
		this._isSelectAllAfterUp = false;
		this._selectedPiecesPosition;
		this._selectedPiecesCharLength;

		this._closingVertexDTh = 4;
		this._closingAngleDTh = 5;

		//------ Animations		
		this._zoomIndex = 0;
		this._zoomTable = [ 1, 1.41, 2 ];
		this._distZoom = 1;
		this._animZoomDiv = 3;
		this._animZoomDth = 0.005;

		this._isModifing = false;
		this._isModified = false;

		this._animModifingDiv = 3;
		this._distPositions = [];
		this._distAngles = [];
		this._distFlips = [];
		this._animFittingDiv = 5;
		this._animBreakingDiv = 7;
		this._animCenteringDiv = 3;
		this._animModifingDth = 0.1;

		this._isSetAheadUndo = false;

		//------ Event handler
		this._scope.view.pause();
		this._tool = new this._scope.Tool();
		this._tool.onMouseDown = this._onMouseDown.bind(this);
		this._tool.onMouseUp = this._onMouseUp.bind(this);
		this._tool.onMouseDrag = this._onMouseDrag.bind(this);
		this._scope.view.onFrame = this._onFrame.bind(this);
		this._scope.view.onResize = this._onResize.bind(this);
	}

	reset(){
		this._reset();
	}

	resize(){
		this._setSize();
	}

	redraw(){
		this._diselectAllPath();
		this._scope.view.draw();
	}
	
	undoRedo(direction){
		this._diselectAllPath();
		this._undoRedo(direction);
	}
	
	rewined(){
		this._diselectAllPath();
		this._rewined();
	}
	
	breakTurbulently(){
		this._diselectAllPath();
		this._breakTurbulently();
	}
	
	centeringPieces(){
		this._centeringPieces();
	}
	
	zoomToggle(){
		this._zoomToggle();
	}
	
	getJtePieceInfoString(){
		return this._getJtePieceInfoString();
	}
	
	getCopiedPieceInfo(){
		var center = this._group.position
		var pieceInfo = this._copyPieceInfo(this._pieceInfo);
		pieceInfo.forEach(function(p){p.position.x -= center.x; p.position.y -= center.y;});
		return pieceInfo;
	}

	getCurrentColoringKey(){
		var colorings = this._patan.colorTheme.getCurrentThemeColorings();
		return colorings['colorings'][this._coloring];
	}

	fixInitPieceInfo(){
		this._setPieceInfoToGroup();
		this._diselectAllPath();
		this._setPiecesStrokeWidth();
		this._scope.view.draw();
	}

	refreshPiecesColor(){
		this._diselectAllPath();
		this._scope.view.draw();
	}
	
	toggleColoring(){
		this._toggleColoring();
		this._diselectAllPath();
		this._scope.view.draw();
	}
	
	applyPieceInfo(pieceInfo){
		this._setSize();
		this._applyPieceInfo(pieceInfo);
		this._diselectAllPath();
		this._setPiecesStrokeWidth();
		this._setUndo(false);
		this._scope.view.draw();
	}

	getDataUrl(){
		this._diselectAllPath();
		this._scope.view.draw();

		var group = this._group.clone();
		var margin = 20 / (this._currZoom * this._zoomRatio);
		var bounds = this._group.bounds;
		var rectangle = new this._scope.Path.Rectangle(
			bounds.x - margin, bounds.y - margin,
			bounds.width + margin * 2, bounds.height + margin * 2
		)
		rectangle.fillColor = new this._scope.Color(0,0);
		rectangle.strokeColor = new this._scope.Color(0,0);
		group.addChild(rectangle);
		var raster = group.rasterize();
		var url = raster.toDataURL();
		raster.remove();
		group.remove();
		return url;
	}


//------ Initial operations
	_reset(){
		this._isAnimation = false;
		this._scope.view.pause();

		this._setSize();
		this._initPiecesGeometry();
		this._setPiecesColor();
		this._setPiecesStrokeWidth();

		this._undoBuffer = [];
		this._setUndo(false);

		this._scope.view.zoom = this._currZoom * this._zoomRatio;
		this._scope.view.draw();
	}

	_setSize(){
		this._scope.view.viewSize = new this._scope.Size(this._patan.width, this._patan.height);
	}


//------------------------------------------------------------------------------
// for internal operation

//------ Internal logics
	_setUndo(isReplaceHistory){
		if( isReplaceHistory ){
			this._undoIndex = ( this._undoIndex - 1 + this._undoSize ) % this._undoSize;
		}
		var positions = [];
		var angles = [];
		var flips = [];
		for( var i = 0; i < this._pieceNum; i++ ){
			positions[i] = this._pieceInfo[i].position;
			angles[i] = this._pieceInfo[i].angle;
			flips[i] = this._pieceInfo[i].flip;
		}
		this._undoBuffer[this._undoIndex] = new this._UndoBuffer(positions, angles, flips);
		this._undoIndex = (this._undoIndex + 1) % this._undoSize;
		if( this._undoStart == this._undoIndex ){
			this._undoStart = (this._undoIndex + 1) % this._undoSize;
		}
		this._undoLimit = Math.min(
			( this._undoIndex - this._undoStart + this._undoSize ) % this._undoSize,
			this._undoSize
		);
	}

	_undoRedo(direction){
		if( this._isAnimation && !this._isExclusiveModifing ){
			return;
		}
		
		var tmpIndex = ( this._undoIndex + direction - 1 + this._undoSize ) % this._undoSize;
		var positions = [];
		var angles = [];
		var flips = [];
		if( tmpIndex == ( this._undoStart - 1 + this._undoSize ) % this._undoSize ||
			tmpIndex == ( this._undoStart + this._undoLimit ) % this._undoSize ){
			return;
		}
		for( var i = 0; i < this._pieceNum; i++ ){
			positions[i] = this._undoBuffer[tmpIndex].positions[i];
			angles[i] = this._undoBuffer[tmpIndex].angles[i];
			flips[i] = this._undoBuffer[tmpIndex].flips[i];
		}
		this._undoIndex = ( tmpIndex + 1 ) % this._undoSize;
		this._movingPiecesAnimation( positions, angles, flips, this._animUndoDiv );
	}

	_rewined(){
		if( this._isAnimation && !this._isExclusiveModifing ){
			return;
		}

		var positions = [];
		var angles = [];
		var flips = [];
		for( var i = 0; i < this._pieceNum; i++ ){
			positions[i] = this._initPositions[i].add( this._scope.view.center.subtract( this._initCenter ) );
			angles[i] = this._initAngles[i];
			flips[i] = this._initFlips[i];
		}
		this._ActivePiecePid = -1;
		this._isSetUndoAfterAnimation = true;
		this._movingPiecesAnimation( positions, angles, flips, this._animRewinedDiv );
	}

	_getJtePieceInfoString(){
		var s = "";
		for( var i = 0; i < this._pieceNum; i++ ){
			s += this._pieceInfo[i].kind.replace(/^_/, '') + ' ';
			s += ( this._pieceInfo[i].position.x - this._group.position.x ) + ' ';
			s += ( this._pieceInfo[i].position.y - this._group.position.y ) + ' ';
			s += ( this._pieceInfo[i].angle / 180 * Math.PI ) + ' ';
			s += this._pieceInfo[i].flip + '\n';
		}
		return s;
	}

	_copyPieceInfo(srcp){
		var dstp = [];
		for( var i = 0; i <srcp.length; i++ ){
			var p = srcp[i];
			dstp.push({
				'kind': p.kind,
				'position': new this._scope.Point(p.position.x, p.position.y),
				'angle': p.angle,
				'flip': p.flip
			});
		}
		return dstp;
	}

	_setPieceInfoToGroup(){
		this._group.removeChildren();
		for( var n = 0; n < this._pieceNum; n++ ){
			this._selectedInfo[n] = false;

			var path;
			path = new this._scope.CompoundPath(this._pathData[this._pieceInfo[n].kind]);
			path.pid = n;
			path.pivot = ([0,0]);
			path.matrix.a = this._pieceInfo[n].flip;
			path.rotate(this._pieceInfo[n].angle);
			path.translate(this._pieceInfo[n].position);
			this._group.addChild(path);

			this._initPositions[n] = this._pieceInfo[n].position;
			this._initAngles[n] = this._pieceInfo[n].angle;
			this._initFlips[n] = this._pieceInfo[n].flip;
		}
		this._initCenter = this._group.position;
		this._translatePiecesInfo(this._scope.view.center.subtract( this._group.position ));
		this._group.position = this._scope.view.center;
		this._group.strokeCap = 'square';
		this._group.strokeJoin = 'bevel';
	}

	_initPiecesGeometry(){
		this._pieceNum = this._patan.pieceset.num;
		this._zoomRatio = this._patan.pieceset.zoomRatio;
		this._pathData = this._patan.pieceset.svgPath;
		this._charLength = this._patan.pieceset.charLen;
		this._pieceInfo = this._copyPieceInfo(this._patan.pieceset.silhouette.p);
		this._setPieceInfoToGroup();
	}

	_applyPieceInfo(src){
		var p = this._copyPieceInfo(src);
		var f = Array(p.length);
		f.fill(true);
		p.forEach(function(pe){
			for( var n = 0; n < this._pieceNum; n++ ){
				if( f[n] && this._pieceInfo[n].kind == pe.kind ){
					f[n] = false;
					this._pieceInfo[n] = pe;
					break;
				}
			}
		}.bind(this));
		this._setPieceInfoToGroup();
	}

	_revertFromPiecesInfo(){
		for( var i = 0; i < this._group.children.length; i++ ){
			var path = this._group.children[i];
			var n = path.pid;
			path.pathData = this._pathData[this._pieceInfo[n].kind];
			path.pivot = ([0,0]);
			path.matrix.a = this._pieceInfo[n].flip;
			path.rotate(this._pieceInfo[n].angle);
			path.translate(this._pieceInfo[n].position);
		}
	}

	_toggleColoring(){
		this._coloring += 1;
		this._coloring %= (this._patan.colorTheme.getCurrentThemeColorings())['variation'];
	}

	_setPiecesColor(){
		var colorings = this._patan.colorTheme.getCurrentThemeColorings();
		this._fills = colorings['fills'][colorings['colorings'][this._coloring]];
		this._strokes = colorings['strokes'][colorings['colorings'][this._coloring]];
		
		for( var i = 0; i < this._group.children.length; i++ ){
			var path = this._group.children[i];
			var n = path.pid;
			var fill = this._fills[n % this._fills.length];
			if( fill == 'none' ){
				path.fillColor = new this._scope.Color(0,0,0,0.01);
			}else{
				path.fillColor = fill;
			}
			var stroke = this._strokes[n % this._strokes.length];
			path.strokeColor = stroke;
		}
	}

	_makeActiveFillColor(color, isRotatable){
		if( isRotatable ){
			color.red = color.red / 2;
			color.green = ( color.green + 0.2 ) / 2;
			color.blue = ( color.blue + 0.6 ) / 2;
		}else{
			color.red = ( color.red + 0.6 ) / 2;
			color.green = ( color.green + 0.2 ) / 2;
			color.blue = color.blue / 2;
		}
		color.alpha = ( color.alpha + 0.5 ) / 2;
		return color;
	}

	_makeSelectedFillColor(color){
		color.red = ( color.red + 0.6 ) / 2;
		color.green = color.green / 2;
		color.blue = ( color.blue + 0.6 ) / 2;
		color.alpha = ( color.alpha + 0.5 ) / 2;
		return color;
	}

	_makeActiveStrokeColor(color){
		color.alpha = 0.75;
		return color;
	}

	_makeSelectedStrokeColor(color){
		return this._makeActiveStrokeColor(color);
	}

	_setPiecesStrokeWidth(){
		this._group.strokeWidth = this._strokeBaseWidth / (this._currZoom * this._zoomRatio);
	}

	_translatePiecesInfo(vector){
		for( var i = 0; i < this._pieceNum; i++ ){
			this._pieceInfo[i].position = this._pieceInfo[i].position.add( vector );
		}
	}


//------------------------------------------------------------------------------
// for mouse operation

//------ Mouse operations

//------ on mouse down
	_onMouseDown(event){
		if( this._isAnimation ){ return; }
		
		this._resetDraged();
		
		var hitResult = this._scope.project.hitTest(event.point, { fill : true });
		if( hitResult && hitResult.item ){					
			this._prepareMovePiece(event, hitResult);
		}else{
			this._prepareSelectRect(event);
		}
	}

	//------ on mouse up
	_onMouseUp(event){
		this._finishSelectRect();

		if( this._activePiecePid >= 0 ){
			this._fixPiece();
		}
	}

	//------ on mouse drag
	_onMouseDrag(event){
		this._judgeDraged(event);

		if( this._activePiecePid >= 0 ){
			this._trail(event);
			this._flip(event);
		}else{
			if( this._isDraged ){
				this._selectRect(event)
			}
		}
	}

	//------ select piece by rectangle
	_prepareSelectRect(event){
		this._rectStartPoint = event.point;
		this._isSelectAllAfterUp = !this._isSelectedPathExist();
		this._diselectAllPath();
	}

	_selectRect(event){
		var rectangle = new this._scope.Rectangle(this._rectStartPoint, event.point);
		this._createSelectRect(rectangle);
		for( var n = 0; n < this._group.children.length; n++ ){
			var path = this._group.children[n];
			var i = path.pid;
			this._selectedInfo[i] = this._pieceInfo[i].position.isInside(rectangle);
		}
		this._setAllSelectedPathColor();
	}

	_finishSelectRect(){
		if( this._activePiecePid < 0 ){
			this._removeSelectRect();
			if( this._isSelectAllAfterUp && !this._isDraged ){
				this._selectAllPath();
			}
		}
		if( this._isSelectedPathExist() ){
			this._lastActivePiecePid = -1;
			this._setSelectedPiecesInfo();
		}
	}

	_setSelectedPiecesInfo(){
		var vector = new this._scope.Point(0,0);
		var surface = 0;
		var tmpGroup = new this._scope.Group();
		for( var i = 0; i < this._pieceNum; i++ ){
			if( !this._selectedInfo[i] ){ continue; }
			var s = Math.pow( this._charLength[this._pieceInfo[i].kind], 2 );
			vector = vector.add( this._pieceInfo[i].position.multiply( s ) );
			surface += s;
		}
		this._selectedPiecesPosition = vector.divide( surface );
		this._selectedPiecesCharLength = Math.sqrt(surface);
	}

	_setAllSelectedPathColor(){
		this._setPiecesColor();
		for( var n = 0; n < this._group.children.length; n++ ){
			var path = this._group.children[n];
			if( this._selectedInfo[path.pid] ){
				path.fillColor = this._makeSelectedFillColor(new this._scope.Color(this._fills[path.pid % this._fills.length]));
				path.strokeColor = this._makeSelectedStrokeColor(new this._scope.Color(this._strokes[path.pid % this._strokes.length]));
			}
		}
	}

	_setAllActivePathColorAndBringToFront(){
		this._setPiecesColor();
		var paths = [];
		for( var n = 0; n < this._group.children.length; n++ ){
			if( this._selectedInfo[this._group.children[n].pid] ){ paths[paths.length] = this._group.children[n]; }
		}
		for( var i = 0; i < paths.length; i++ ){
			var path = paths[i];
			path.bringToFront();
			path.fillColor = this._makeActiveFillColor(new this._scope.Color(this._fills[path.pid % this._fills.length]), (this._ratiodtheta > 0));
			path.strokeColor = this._makeActiveStrokeColor(new this._scope.Color(this._strokes[path.pid % this._strokes.length]));
		}
	}

	_selectAllPath(){
		for( var i = 0; i < this._pieceNum; i++ ){
			this._selectedInfo[i] = true;
		}
		this._setAllSelectedPathColor();
	}

	_diselectAllPath(){
		this._setPiecesColor();
		for( var i = 0; i < this._pieceNum; i++ ){
			this._selectedInfo[i] = false;
		}
	}

	_isSelectedPathExist(){
		for( var i = 0; i < this._pieceNum; i++ ){
			if( this._selectedInfo[i] ){
				return true;
			}
		}
		return false;
	}

	_removeSelectRect(){
		if( this._edgeGroup != null ){
			this._edgeGroup.remove();
		}
	}

	_createSelectRect(rectangle){
		this._removeSelectRect();
		this._edgeGroup = new this._scope.Group();
		var rect = new this._scope.Path.Rectangle(rectangle);
		rect.strokeColor = new this._scope.Color(1,1,1,0.5);
		rect.strokeWidth = 3 / (this._currZoom * this._zoomRatio);
		this._edgeGroup.addChild(rect);
		rect = new this._scope.Path.Rectangle(rectangle);
		rect.strokeColor = new this._scope.Color(0,0,0,1);
		rect.strokeWidth = 1 / (this._currZoom * this._zoomRatio);
		this._edgeGroup.addChild(rect);
	}

	_resetDraged(){
		this._isDraged = false;
		this._isDragedMore = false;
		this._dragAccumAbsDelta = new this._scope.Point(0,0);
	}

	_judgeDraged(event){
		if( !this._isDragedMore ){
			this._dragAccumAbsDelta = this._dragAccumAbsDelta + event.delta.abs();
			this._isDraged = (this._dragAccumAbsDelta.length > this._dragAccumAbsDth);
			this._isDragedMore = (this._dragAccumAbsDelta.length > this._dragMoreAccumAbsDth);
		}
	}

	//------ move piece locaion
	_resetFlip(){
		this._flipDeltaIndex = 0;
		this._flipAccumDelta = new this._scope.Point(0,0);
		this._flipAccumAbsDelta = new this._scope.Point(0,0);
		for( var i = 0; i < this._flipAccumLength; i++ ){ this._flipDeltas[i] = new this._scope.Point(0,0); }
	}

	_prepareMovePiece(event, hitResult){
		var path = hitResult.item;
		this._activePiecePid = path.pid;

		this._resetFlip();

		if( this._selectedInfo[this._activePiecePid] ){
			// one or more pieces selected
			this._calcTrailRatioDTheta(event, this._selectedPiecesPosition, this._selectedPiecesCharLength);
			this._setAllActivePathColorAndBringToFront();
		}
		else{
			// mono piece selected
			path.bringToFront();
			this._calcTrailRatioDTheta(event, this._pieceInfo[this._activePiecePid].position, this._charLength[this._pieceInfo[this._activePiecePid].kind]);
			this._diselectAllPath();
			path.fillColor = this._makeActiveFillColor(new this._scope.Color(this._fills[this._activePiecePid % this._fills.length]), (this._ratiodtheta > 0));
			path.strokeColor = this._makeActiveStrokeColor(new this._scope.Color(this._strokes[this._activePiecePid % this._strokes.length]));
		}
	}

	_calcTrailRatioDTheta(event, position, charlen){
		var vector = event.point.subtract( position );
		this._ratiodtheta = vector.length / charlen;
		this._ratiodtheta = (this._ratiodtheta < this._ratiodthetamin) ? 0 : ( this._ratiodthetamin + (this._ratiodtheta - this._ratiodthetamin) / this._ratiodthetath );
		this._ratiodtheta = Math.min( this._ratiodtheta, 1 );
	}

	_calcTrailDAngle(vector, prevec){
		var dAngle = prevec.getAngle( vector );
		dAngle = isNaN( dAngle ) ? 0 : dAngle;
		dAngle *= this._ratiodtheta;
		dAngle *= (prevec.cross( vector ) < 0) ? -1 : 1;
		return dAngle;
	}

	_trail(event){
		var vector;
		var prevec;
		var dAngle;
		var position;

		if( this._selectedInfo[this._activePiecePid] ){
			// one or more pieces selected
			vector = event.point.subtract( this._selectedPiecesPosition );
			prevec = event.lastPoint.subtract( this._selectedPiecesPosition );
			dAngle = this._calcTrailDAngle( vector, prevec );
			position = event.point.subtract( prevec.rotate(dAngle) );
			for( var n = 0; n < this._group.children.length; n++ ){
				var path = this._group.children[n];
				var i = path.pid;
				if( !this._selectedInfo[i] ){ continue; }

				var piecePosition = position.add( (this._pieceInfo[i].position.subtract( this._selectedPiecesPosition )).rotate(dAngle) );
				path.pivot = this._pieceInfo[i].position;
				path.rotate((dAngle + 360)%360);
				path.position = path.position.add( piecePosition.subtract( this._pieceInfo[i].position ) );

				this._pieceInfo[i].position = piecePosition;
				this._pieceInfo[i].angle = (this._pieceInfo[i].angle + dAngle + 360) % 360; 
			}
			this._selectedPiecesPosition = position;
		}
		else{
			// mono piece selected
			vector = event.point.subtract( this._pieceInfo[this._activePiecePid].position );
			prevec = event.lastPoint.subtract( this._pieceInfo[this._activePiecePid].position );
			dAngle = this._calcTrailDAngle( vector, prevec );

			position = event.point.subtract( prevec.rotate(dAngle) );
			this._group.lastChild.pivot = this._pieceInfo[this._activePiecePid].position;
			this._group.lastChild.rotate((dAngle + 360)%360);
			this._group.lastChild.position = this._group.lastChild.position.add( position.subtract( this._pieceInfo[this._activePiecePid].position ) );

			this._pieceInfo[this._activePiecePid].angle += dAngle + 360;
			this._pieceInfo[this._activePiecePid].angle %= 360;
			this._pieceInfo[this._activePiecePid].position = position;
		}
	}

	_flip(event){
		var vector;
		this._flipAccumDelta = this._flipAccumDelta.add( event.delta.subtract( this._flipDeltas[this._flipDeltaIndex] ) );
		this._flipAccumAbsDelta = this._flipAccumAbsDelta.add( event.delta.abs().subtract( this._flipDeltas[this._flipDeltaIndex].abs() ) );
		this._flipDeltas[this._flipDeltaIndex] = event.delta;
		this._flipDeltaIndex = (this._flipDeltaIndex + 1) % this._flipAccumLength;
		if( this._flipAccumDelta.abs().x < this._flipAbsDeltaMin &&
			this._flipAccumAbsDelta.x > this._flipAbsDeltaMax &&
			this._flipAccumAbsDelta.y < this._flipAbsDeltaMin
		){
			for( var i = 0; i < this._pieceNum; i++ ){
				if( !this._selectedInfo[i] && i != this._activePiecePid ){ continue; }
				this._pieceInfo[i].flip = -this._pieceInfo[i].flip;
				this._pieceInfo[i].angle = ( 360 - this._pieceInfo[i].angle + 360 ) % 360;
				vector = event.point.subtract( this._pieceInfo[i].position );
				vector.x = -vector.x;
				this._pieceInfo[i].position = event.point.subtract( vector );
			}
			this._resetFlip();
			this._revertFromPiecesInfo();
		}
	}

	//------ adjust piece location
	_fixPiece(){
		this._setPiecesColor();
		this._setAllSelectedPathColor();

		if( !this._closingVertex() ){
			this._setUndo( this._activePiecePid == this._lastActivePiecePid && !this._isDragedMore );
			this._lastActivePiecePid = this._activePiecePid;
			this._activePiecePid = -1;
		}
	}

	_closingVertex(){
		if( this._activePiecePid < 0 ){ return false; }

		var isExecuteClosing = false;

		//--- search nearest point
		var vector;
		var pivotPoint;
		var dAngle = 0;
		var closingPiecePid = -1;
		var nearestPiecePid = -1;
		var shortestVector = new this._scope.Point(this._closingVertexDTh, 0);
		var activePath, path;

		// vertex to vertex
		for( var n = 0; n < this._group.children.length; n++ ){
			if( this._group.children[n].pid != this._activePiecePid && !this._selectedInfo[this._group.children[n].pid] ){ continue; }
			activePath = this._group.children[n];

			for( var i = 0; i < this._group.children.length; i++ ){
				if( this._group.children[i].pid == this._activePiecePid || this._selectedInfo[this._group.children[i].pid] ){ continue; }
				path = this._group.children[i];
				
				var diff = this._pieceInfo[activePath.pid].position.getDistance(this._pieceInfo[path.pid].position);
				if( diff > this._charLength[this._pieceInfo[activePath.pid].kind]*2 + this._charLength[this._pieceInfo[path.pid].kind]*2){
					//console.log("*");
					continue; }

				for( var j = 0; j < activePath.lastChild.segments.length; j++ ){
					for( var k = 0; k < path.lastChild.segments.length; k++ ){
						vector = path.lastChild.segments[k].point.subtract( activePath.lastChild.segments[j].point );
						if( vector.length < shortestVector.length ){
							pivotPoint = activePath.lastChild.segments[j].point;
							shortestVector = new this._scope.Point(vector);
							closingPiecePid = activePath.pid;
							nearestPiecePid = path.pid;
						}
					}
				}
			}
		}		

		// vertex to path
		if( nearestPiecePid < 0 ){
			for( var n = 0; n < this._group.children.length; n++ ){
				if( this._group.children[n].pid != this._activePiecePid && !this._selectedInfo[this._group.children[n].pid] ){ continue; }
				activePath = this._group.children[n];

				for( var i = 0; i < this._group.children.length; i++ ){
					if( this._group.children[i].pid == this._activePiecePid || this._selectedInfo[this._group.children[i].pid] ){ continue; }
					path = this._group.children[i];

					var diff = this._pieceInfo[activePath.pid].position.getDistance(this._pieceInfo[path.pid].position);
					if( diff > this._charLength[this._pieceInfo[activePath.pid].kind]*2 + this._charLength[this._pieceInfo[path.pid].kind]*2){
						//console.log("*");
						continue; }
			
					for( var j = 0; j < activePath.lastChild.segments.length; j++ ){
						vector = path.lastChild.getNearestPoint( activePath.lastChild.segments[j].point ).subtract( activePath.lastChild.segments[j].point );
						if( vector.length < shortestVector.length ){
							pivotPoint = activePath.lastChild.segments[j].point.add( vector );
							shortestVector = new this._scope.Point(vector);
							closingPiecePid = activePath.pid;
							nearestPiecePid = path.pid;
						}
					}
				}
			}
		}

		// angle to angle
		if( nearestPiecePid >= 0 && !this._isDraged ){
			var adjustAngle = this._pieceInfo[closingPiecePid].angle - this._pieceInfo[nearestPiecePid].angle;
			var modAdjustAngle = adjustAngle % 15;
			dAngle = modAdjustAngle - ( (Math.abs(modAdjustAngle) < 7.5)? 0 : ((modAdjustAngle > 0)? 15 : -15) );
			if( Math.abs(dAngle) > this._closingAngleDTh ){
				dAngle = 0;
			}else{
				dAngle = (360 - dAngle) % 360;
			}
		}

		//--- closing
		var positions = [];
		var angles = [];
		var flips = [];
		for( var i = 0; i < this._pieceNum; i++ ){
			positions[i] = this._pieceInfo[i].position;
			angles[i] = this._pieceInfo[i].angle;
			flips[i] = this._pieceInfo[i].flip;
		}

		if( nearestPiecePid >= 0 ){
			if( shortestVector.length > 0 || dAngle != 0 ){

				for( var i = 0; i < this._pieceNum; i++ ){
					if( i != this._activePiecePid && !this._selectedInfo[i] ){ continue; }
					
					// vertex closing
					positions[i] = positions[i].add( shortestVector );

					// adjust angle slice
					angles[i] += dAngle;
					angles[i] %= 360;
					vector = pivotPoint.subtract( positions[i] );
					positions[i] = positions[i].add( vector.subtract( vector.rotate(dAngle) ) );
				}
				isExecuteClosing = true;
			}
		}
		else if( !this._isDraged ){

			//--- angle round 15 degree
			var modAngle = 0;
			for( var i = 0; i < this._pieceNum; i++ ){
				if( i != this._activePiecePid && !this._selectedInfo[i] ){ continue; }
				dAngle = this._pieceInfo[this._activePiecePid].angle % 15;
				dAngle = ( ( dAngle < 7.5 )? 0 : 15 ) - dAngle;
				if( Math.abs(dAngle) < Math.abs(modAngle) || modAngle == 0 ){
					modAngle = dAngle;
				}
			}
			if( modAngle != 0 ){
				if( this._selectedInfo[this._activePiecePid] ){
					// one or more pieces selected
					for( var i = 0; i < this._pieceNum; i++ ){
						if( !this._selectedInfo[i] ){ continue; }
						positions[i] = this._selectedPiecesPosition.add( (positions[i].subtract( this._selectedPiecesPosition )).rotate(modAngle) );
						angles[i] += modAngle;
					}
				}else{
					// mono piece selected
					angles[this._activePiecePid] += modAngle;
				}
				isExecuteClosing = true;
			}
		}

		// animation
		if( isExecuteClosing ){
			this._isSetUndoAfterAnimation = true;
			this._movingPiecesAnimation( positions, angles, flips, this._animClosingDiv, !this._isDragedMore );
		}
		return isExecuteClosing;
	}


//------------------------------------------------------------------------------
// for animation

//------ Animations		

	_movingPiecesAnimation(positions, angles, flips, div, isReplaceUndo ){
		this._isModifing = true;
		this._isModified = false;
		this._isExclusiveModifing = true;
		this._isSetAheadUndo = !isReplaceUndo;
		this._distPositions = positions;
		this._distAngles = angles;
		this._distFlips = flips;
		this._animModifingDiv = div;
		this._animationOn();
	}

	_breakTurbulently(){
		if( this._isExclusiveModifing ){
			return;
		}
		
		for( var i = 0; i < this._pieceNum; i++ ){
			var vector = this._group.position.subtract( this._pieceInfo[i].position );
			var dAngle = 10 + Math.random() * 100 / ( 1 + vector.length / 200 );
			dAngle *= ( Math.random() < 0.5 )? 1 : -1;
			vector.angle += dAngle;
			vector = vector.multiply( 0.5 + Math.random() * 1.5 );
			this._distPositions[i] = this._pieceInfo[i].position.add( vector );
			this._distAngles[i] = this._pieceInfo[i].angle + dAngle * 3 * Math.random();
			this._distFlips[i] = (Math.random() < 0.5)? 1 : -1;
		}
		this._isModifing = true;
		this._isSetUndoAfterAnimation = true;
		this._activePiecePid = -1;
		this._animModifingDiv = this._animBreakingDiv;
		this._animationOn();
	}

	_centeringPieces(){
		if( this._isExclusiveModifing ){
			return;
		}
		
		var vector = this._scope.view.center.subtract( this._group.position );
		for( var i = 0; i < this._pieceNum; i++ ){
			this._distPositions[i] = this._pieceInfo[i].position.add( vector );
			if( !this._isModifing ){
				this._distAngles[i] = this._pieceInfo[i].angle;
				this._distFlips[i] = this._pieceInfo[i].flip;
			}
		}
		this._isModifing = true;
		this._animModifingDiv = this._animCenteringDiv;
		this._animationOn();
	}

	_zoomToggle(){
		this._zoomIndex = ( this._zoomIndex + 1 ) % this._zoomTable.length;
		this._distZoom = this._zoomTable[this._zoomIndex];
		this._animationOn();
	}

	_animationOn(){
		if( this._isAnimtaion ){
			return;
		}
		
		this._isAnimation = true;
		this._scope.view.play();
	}

	_onFrame(event){
		var isTerm = true;

		if( Math.abs( this._distZoom - this._currZoom ) > this._animZoomDth ){
			isTerm = false;
			this._currZoom += ( this._distZoom - this._currZoom ) / this._animZoomDiv;
			this._setPiecesStrokeWidth();
			this._scope.view.zoom = this._currZoom * this._zoomRatio;
		}

		if( this._isModifing ){
			var isTermModifing = true;
			for( var i = 0; i < this._pieceNum; i++ ){
				var vector = this._distPositions[i].subtract( this._pieceInfo[i].position );
				var	dAngle = this._distAngles[i] - this._pieceInfo[i].angle;
				if( Math.abs(dAngle) > 180 ){
					dAngle = ( 360 - Math.abs(dAngle) )*( (dAngle > 0)? -1 : 1);
				}
				var dFlip = this._distFlips[i] - this._pieceInfo[i].flip;
				if( vector.length > this._animModifingDth ||
					Math.abs(dAngle) > this._animModifingDth ||
					Math.abs(dFlip) > this._animModifingDth
				){
					isTermModifing = false;
					isTerm = false;
					var division;

					division = 1 + ( vector.length - this._animModifingDth );
					division = Math.min( division, this._animModifingDiv );
					division = Math.max( division, 1 );
					vector = vector.divide( division );
					this._pieceInfo[i].position = this._pieceInfo[i].position.add( vector );

					division = 1 + ( Math.abs(dAngle) - this._animModifingDth );
					division = Math.min( division, this._animModifingDiv );
					division = Math.max( division, 1 );
					dAngle /= division;
					this._pieceInfo[i].angle = ( 360 + this._pieceInfo[i].angle + dAngle ) % 360;

					division = 1 + ( Math.abs(dFlip) - this._animModifingDth );
					division = Math.min( division, this._animModifingDiv );
					division = Math.max( division, 1 );
					dFlip /= division;
					this._pieceInfo[i].flip += dFlip;
				}
			}
			if( !isTermModifing ){
				this._isModified = true;
				this._revertFromPiecesInfo();
			}else{
				this._isModifing = false;
				this._isExclusiveModifing = false;
				for( var i = 0; i < this._pieceNum; i++ ){
					this._pieceInfo[i].position = this._distPositions[i];
					this._pieceInfo[i].angle = this._distAngles[i];
					this._pieceInfo[i].flip = this._distFlips[i];
				}
				if( this._isSetUndoAfterAnimation && this._isModified ){
					this._setUndo( this._activePiecePid == this._lastActivePiecePid && this._activePiecePid >= 0 && !this._isSetAheadUndo );
					this._isForceSetUndo = false;
				}
				this._isSetUndoAfterAnimation = false;
				this._lastActivePiecePid = this._activePiecePid;
				this._activePiecePid = -1;
			}
		}
		
		if( isTerm ){
			this._isAnimation = false;
			this._scope.view.pause();
			this._revertFromPiecesInfo();
		}
	}

	_onResize(event){
		if( this._group != null ){
			this._translatePiecesInfo(this._scope.view.center.subtract( this._group.position ));
			this._group.position = this._scope.view.center;
		}
	}

}
