// patan logo canvas

class LogoCanvas{
	constructor(patan, element){
		this._patan = patan;
		this._scope = new paper.PaperScope();
		this._scope.setup(element);

		this._group = new this._scope.Group();
	}

	reset(){
		this._reset();
	}

	refreshPiecesColor(){
		this._setPiecesColor();
		this._scope.view.draw();
	}

	_setPiecesColor(){
		this._group.fillColor = this._patan.colorTheme.getCurrentThemeColor('cid_logofill');
		this._group.strokeColor = this._patan.colorTheme.getCurrentThemeColor('cid_logostroke');
	}

	_reset(){
		var scope = this._scope;
		var group = this._group;
		
		var pathData = this._patan.pieceset.svgPath;
		var pieceInfo = this._patan.pieceset.silhouette.p;
		var zoomRatio = this._patan.pieceset.zoomRatio;

		scope.view.pause();
		scope.view.viewSize = new scope.Size(400,300);

		group.removeChildren();
		for( var n = 0; n < pieceInfo.length; n++ ){
			var cpath = new scope.CompoundPath(pathData[pieceInfo[n].kind]);
			cpath.pivot = ([0,0]);
			cpath.matrix.a = pieceInfo[n].flip;
			cpath.rotate(pieceInfo[n].angle);
			cpath.translate(pieceInfo[n].position);
			
			var newcpath = getHandWritingEffect(cpath,1.5);
			cpath.remove();
			newcpath.pivot = ([0,0]);
			
			group.addChild(newcpath);
		}
		group.rotate(20);
		group.scaling = zoomRatio;
		group.position = scope.view.center;
		group.strokeCap = 'round';
		group.strokeJoin = 'bevel';
		group.strokeWidth = 1.5;
		this._setPiecesColor();

		scope.view.draw();
	}
}
