// patan work canvas

class WorkCanvas{
	constructor(element){
		this._scope = new paper.PaperScope();;
		this._scope.setup(element);
	}

	get scope(){ return this._scope; }

	getDataURL(svgPath, zoomRatio, pieceInfo, fill, stroke, zoom){
		var scope = this._scope;
		scope.activate();
		var group = new scope.Group();

		for( var n = 0; n < pieceInfo.length; n++ ){
			var cpath = this._getCpath(svgPath[pieceInfo[n].kind], pieceInfo[n],
				fill, stroke);
			group.addChild(cpath);
		}
		group.strokeWidth = 4.0 * (zoom * zoomRatio);
		group.scaling = zoom * zoomRatio;
		
		var bbox = group.bounds;
		scope.view.viewSize = new scope.Size(bbox.width + 4, bbox.height + 4);
		group.position = scope.view.center;

		scope.view.draw();
		var raster = group.rasterize();
		var url = raster.toDataURL();
		var size = scope.view.viewSize.clone();

		raster.remove();
		group.remove();

		return {
			'url': url,
			'width': size.width,
			'height': size.height,
		};
	}

	getDataURL_wide(pieceset, fill, stroke, pieceInfo, fills, strokes){
		var scope = this._scope;
		scope.activate();
		var layer = new this._scope.Layer();
		var group1 = new this._scope.Group();
		var group2 = new this._scope.Group();
		layer.addChild(group1);
		layer.addChild(group2);

		var zoomRatio = pieceset.zoomRatio;
		var svgPath = pieceset.svgPath;

		for( var n = 0; n < pieceInfo.length; n++ ){
			var p = pieceset.silhouette.p;
			var cpath = this._getCpath(svgPath[p[n].kind], p[n],
				fill, stroke);
			var newcpath = getHandWritingEffect(cpath, 1.5);
			cpath.remove();
			if( fill != 'none' ){
				newcpath.fillColor = fill;
			}
			newcpath.strokeColor = stroke;
			newcpath.strokeCap = 'round';
			newcpath.strokeJoin = 'bevel';
			group1.addChild(newcpath);

			p = pieceInfo;
			cpath = this._getCpath(svgPath[p[n].kind], p[n],
				fills[n % fills.length], strokes[n % strokes.length]);
			group2.addChild(cpath);
		}

		var zoom = 0.66;
		group1.scaling = zoom * zoomRatio;
		group1.strokeWidth = 1.5 / (zoom * zoomRatio);
		var bbox1 = group1.bounds;

		zoom = 1;
		group2.scaling = zoom * zoomRatio;
		group2.strokeWidth = 1.5 / (zoom * zoomRatio);
		var bbox2 = group2.bounds;

		var paddingh = 40;
		var paddingv = 40;
		group1.position = new scope.Point(0, 0);
		group2.position = new scope.Point(0, 0);
		group2.position.x += (bbox1.width + bbox2.width) / 2 + paddingh * 2;

		var bbox = layer.bounds;
		var size = new scope.Size(bbox.width + paddingh * 4, bbox.height + paddingv * 2);

		if( size.width / size.height > 16/9 ){
			size.height = size.width / 16 * 9;
		}else{
			size.width = size.height / 9 * 16;
		}

		var isResize = bbox1.height < size.height / 3;
		if( isResize ){
			zoom = (size.height / 3) / bbox1.height;
			//paddingh *= zoom;
			group2.scaling = 1 / zoom;
			bbox2 = group2.bounds;
			group2.position = new scope.Point(0, 0);
			group2.position.x += (bbox1.width + bbox2.width) / 2 + paddingh * 2;
			bbox = layer.bounds;
			size = new scope.Size(bbox.width + paddingh * 4, bbox.height + paddingv * 2);
		
		}

		if( size.width / size.height > 16/9 ){
			size.height = size.width / 16 * 9;
		}else{
			size.width = size.height / 9 * 16;
		}

		layer.position = scope.view.center;

		var rect = new scope.Path.Rectangle(0, 0, size.width, size.height);
		rect.position = scope.view.center;
		layer.addChild(rect);
		layer.scaling = 480 / size.height;

		size = new scope.Size(rect.size);
		var raster = layer.rasterize();
		var url = raster.toDataURL();

		raster.remove();
		layer.remove();

		return {
			'url': url,
			'width': size.width,
			'height': size.height,
		};
	}

	_getCpath(path, p, fill, stroke){
		var cpath = new this._scope.CompoundPath(path);
		cpath.pivot = ([0,0]);
		cpath.matrix.a = p.flip;
		cpath.rotate(p.angle);
		cpath.translate(p.position);
		if( fill != 'none' ){
			cpath.fillColor = fill;
		}
		cpath.strokeColor = stroke;
		cpath.strokeCap = 'round';
		cpath.strokeJoin = 'bevel';

		return cpath;
	}
}
