// jte library

class Jte{
	constructor(scope){
		this._scope = scope;
		this._surfaceNormalize = 160 * 160;
		this._nodeNum = 100;
		this._epochNum = 10000;
		this._learningRate = 0.1;
	}

	// json reviver
	jsonReviver(key, value){
		if( key == "position" && value[0] == "Point" ){
			return new this._scope.Point(value[1], value[2]);
		}
		return value;
	}

	// parse x y positions
	parsePos(s,ox,oy){
		var t = s.split(/\s+/);
		var v = [];
		for( var i = 0; i < t.length/2; i++ ){
			v[i*2] = Number(t[i*2])-ox;
			v[i*2+1] = Number(t[i*2+1])-oy;
		}
		return v;
	}

	// name and tags
	separateTags(s){
		var t = s.split(/\s*#/);
		var name = t[0];
		t.shift();
		return {'name': name, 'tags': t};
	}

	jointTags(name, tags){
		var s = name;
		tags.forEach(function(t){s += ' #' + t;});
		return s;
	}

	// expand silhouette from xml content
	_expandSilhouette(obj){
		// geometrics
		var pieceInfo = [];
		var t = obj.text().replace(/^[\n\t\s]+|[\n\t\s]+$/g, '').split(/[\n\t\s]+/);
		for( var i = 0; i < t.length/5; i++){
			var p = {
				'kind': t[i*5],
				'position': new this._scope.Point(Number(t[i*5+1]), Number(t[i*5+2])),
				'angle': Number(t[i*5+3]) * 180 / Math.PI,
				'flip': Number(t[i*5+4]),
			};
			pieceInfo.push(p);
		}

		// name, tags (#...)
		t = this.separateTags(obj.attr('name'));

		var silhouette = {
			'name': t['name'],
			'tags': t['tags'],
			'p': pieceInfo,
		};

		return silhouette;
	}

	// expand pieceset & silhouettes from jte xml
	expandInitialJteXml(silhouettePuzzle){
		// initial jte xml(tangram)
		var initialJteXml =
		`<?xml version="1.0" encoding="UTF-8" ?>
		<jte format="0.0">
			<lastupdate date="no data"/>
			<pieceset name="Tangram" num="7">
				<geom id="Mt">
					<p kind="origin">28.0 28.0</p>
					<p kind="moveTo">0.0 0.0</p>
					<p kind="lineTo">0.0 80.0</p>
					<p kind="lineTo">80.0 0.0</p>
					<p kind="lineTo">0.0 0.0</p>
				</geom>
				<geom id="St">
					<p kind="origin">40.0 14.0</p>
					<p kind="moveTo">0.0 0.0</p>
					<p kind="lineTo">40.0 40.0</p>
					<p kind="lineTo">80.0 0.0</p>
					<p kind="lineTo">0.0 0.0</p>
				</geom>
				<geom id="Sq">
					<p kind="origin">40.0 40.0</p>
					<p kind="moveTo">40.0 0.0</p>
					<p kind="lineTo">0.0 40.0</p>
					<p kind="lineTo">40.0 80.0</p>
					<p kind="lineTo">80.0 40.0</p>
					<p kind="lineTo">40.0 0.0</p>
				</geom>
				<geom id="Pa">
					<p kind="origin">20.0 60.0</p>
					<p kind="moveTo">0.0 40.0</p>
					<p kind="lineTo">0.0 120.0</p>
					<p kind="lineTo">40.0 80.0</p>
					<p kind="lineTo">40.0 0.0</p>
					<p kind="lineTo">0.0 40.0</p>
				</geom>
				<geom id="Lt">
					<p kind="origin">56.0 80.0</p>
					<p kind="moveTo">0.0 80.0</p>
					<p kind="lineTo">80.0 160.0</p>
					<p kind="lineTo">80.0 0.0</p>
					<p kind="lineTo">0.0 80.0</p>
				</geom>
				<silhouette name="Anker puzzle no.8">
					Mt -52.0000005756408 -51.99999955748767 6.283185307179586 1.0
					St 39.9999994243592 -65.9999995574877 6.283185307179586 1.0
					Sq -5.756408034685023E-7 -39.99999955748764 0.0 -1.0
					St -26.00000057564084 4.425123503182249E-7 4.71238898038469 1.0
					Pa -60.000000575640776 20.000000442512345 6.283185307179586 1.0
					Lt 55.99999942435921 4.425123290019428E-7 6.283185307179586 1.0
					Lt -5.756407785995066E-7 56.00000044251237 1.5707963267948966 1.0
				</silhouette>
			</pieceset>
		</jte>`;
		return this.expandJteXml(initialJteXml,silhouettePuzzle);
	}

	expandJteXml(xml,silhouettePuzzle,withSilhouettes=true){
		var obj;
		var xmlDoc = null;
		try{
			xmlDoc = $.parseXML(xml);
		}catch(e){
			return {'result': false, 'message': 'not xml'};
		}
		// check jte xml
		obj = $(xmlDoc).find('jte');
		if( obj.length == 0 ){
			return {'result': false, 'message': 'not jte'};
		}
		// check jte format version
		if( obj.eq(0).attr('format') != '0.0' ){
			return {'result': false, 'message': 'not supported jte format version'};
		}

		// check pieceset
		obj = $(xmlDoc).find('pieceset');
		if( obj.length == 0 ){
			return {'result': false, 'message': 'pieceset not found'};
		}

		// expand pieceset
		var tmpName = obj.eq(0).attr('name');
		var tmpNum = Number(obj.eq(0).attr('num'));
		var tmpSvgPath = {};
		var tmpCharLen = {};
		var tmpSilhouette = {};

		var areas = {};
		obj = $(xmlDoc).find('pieceset').eq(0).find('geom');
		for( var i = 0; i < obj.length; i++){
			var pobj = obj.eq(i).find('p');
			var svgPath = '';
			var ox = 0;
			var oy = 0;
			var v = [];
			var kind = obj.eq(i).attr('id');
			for( var j = 0; j < pobj.length; j++){
				if( pobj.eq(j).attr('kind') == 'origin' ){
					v = this.parsePos( pobj.eq(j).text(), 0, 0);
					ox = v[0];
					oy = v[1];
				}else if( pobj.eq(j).attr('kind') == 'moveTo' ){
					v =  this.parsePos( pobj.eq(j).text(), ox, oy);
					svgPath += 'M '+v[0]+' '+v[1]+' ';
				}else if( pobj.eq(j).attr('kind') == 'lineTo' ){
					v =  this.parsePos( pobj.eq(j).text(), ox, oy);
					svgPath += 'L '+v[0]+' '+v[1]+' ';
				}else if( pobj.eq(j).attr('kind') == 'curveTo' ){
					v =  this.parsePos( pobj.eq(j).text(), ox, oy);
					svgPath += 'C '+v[0]+' '+v[1]+' '+v[2]+' '+v[3]+' '+v[4]+' '+v[5]+' ';
				}
			}
			svgPath += 'Z';
			tmpSvgPath[kind] = svgPath;
			
			var path = new this._scope.Path(svgPath);
			var area = Math.abs(path.area);
			path.remove();
			areas[kind] = area;
			tmpCharLen[kind] = Math.sqrt(area);
		}

		obj = $(xmlDoc).find('pieceset').eq(0).find('silhouette');
		if( obj.length == 0 ){
			return {'result': false, 'message': 'pieceset has no silhouette'};
		}
		tmpSilhouette = this._expandSilhouette(obj.eq(0));

		var area = 0;
		for( var i = 0; i < tmpSilhouette['p'].length; i++ ){
			area += areas[tmpSilhouette['p'][i]['kind']];
		}

		var pieceset = {
			'name': tmpName,
			'num': tmpNum,
			'svgPath': tmpSvgPath,
			'charLen': tmpCharLen,
			'silhouette': tmpSilhouette,
			'zoomRatio': 1 / Math.sqrt(area / Math.pow(160,2)),
		};
		silhouettePuzzle.pieceset = pieceset;

		// expand silhouettes
		var silhouettes = [];
		if( withSilhouettes ){
			obj = $(xmlDoc).find('silhouettes').eq(0).find('silhouette');
			for( var i = 0; i < obj.length; i++){
				tmpSilhouette =this. _expandSilhouette(obj.eq(i));
				silhouettes[i] = tmpSilhouette;
			}
		}
		silhouettePuzzle.silhouettes = silhouettes;

		return {'result': true, 'message': ''};
	}

	buildJteXml(pieceset, silhouettes){
		var jte =
		'<?xml version="1.0" encoding="UTF-8" ?>\n' +
		'<jte format="0.0">\n';

		// pieceset
		jte += '\t<pieceset name="' + pieceset.name +
			'" num="' + pieceset.silhouette.p.length + '">\n';
			
		var kinds = Object.keys(pieceset.svgPath);
		kinds.forEach(function(kind){
			var t = pieceset.svgPath[kind].split(/\s+/);
			jte += '\t\t<geom id="' + kind + '">\n';
			jte += '\t\t\t<p kind="origin">0 0</p>\n';

			for( var i = 0; i < t.length; i++ ){
				if( t[i].length > 1 ){
					t.splice(i+1, 0, t[i].slice(1));
					t[i] = t[i].slice(0, 1);
				}
				if( t[i] == 'M' ){
					jte += '\t\t\t<p kind="moveTo">' +
						t[i+1]+' '+t[i+2]+'</p>\n';
						i += 2;
				}else if( t[i] == 'L' ){
					jte += '\t\t\t<p kind="lineTo">' +
						t[i+1]+' '+t[i+2]+'</p>\n';
						i += 2;
				}else if( t[i] == 'C' ){
					jte += '\t\t\t<p kind="curveTo">' +
						t[i+1]+' '+t[i+2]+' ' +
						t[i+3]+' '+t[i+4]+' ' +
						t[i+5]+' '+t[i+6]+'</p>\n';
						i += 6;
				}
			}

			jte += '\t\t</geom>\n';
		});

		jte += this._getJtePieceInfoString({
			name: pieceset.silhouette.name,
			tags: [],
			p: pieceset.silhouette.p,
		});
		jte += '\t</pieceset>\n';

		// silhouettes
		jte += '\t<silhouettes>\n';
		silhouettes.forEach(function(silhouette){jte += this._getJtePieceInfoString(silhouette);}.bind(this));
		jte += '\t</silhouettes>\n';
		jte += '</jte>\n';

		return jte;
	}

	_getJtePieceInfoString(silhouette){
		var s = '\t\t<silhouette name="' + jointTags(silhouette) + '">\n';
		silhouette.p.forEach(function(p){
			s += '\t\t\t'
			s += p.kind + ' ';
			s += p.position.x + ' ';
			s += p.position.y + ' ';
			s += ( p.angle / 180 * Math.PI ) + ' ';
			s += p.flip + '\n';
		});
		s += '\t\t</silhouette>\n';
		return s;
	}

	getFeatureVector(pieceset, silhouette){
		var scope = this._scope;
		scope.activate();

		var groupm = new scope.Group();
		silhouette.p.forEach(function(p){
			var cpath = new scope.CompoundPath(pieceset.svgPath[p.kind]);
			cpath.pivot = ([0,0]);
			cpath.matrix.a = p.flip;
			cpath.rotate(p.angle);
			cpath.translate(p.position);
			cpath.fillColor = 'red';
			cpath.strokeColor = 'red';
			cpath.strokeCap = 'round';
			cpath.strokeJoin = 'bevel';
			groupm.addChild(cpath);
		});
		var feature;
		feature = this._getFeature(groupm, 0.2, pieceset.zoomRatio, 9.0, true);
		feature.reducedEdgeScore = this._getFeature(groupm, 0.02, pieceset.zoomRatio, 12.2, false).edgeScore;
		groupm.remove();

		return feature;
	}

	_getFeature(groupm, zoom, zoomRatio, normalize, isCalcCoordDev){
		var scope = this._scope;
		var group = groupm.clone();
		group.strokeWidth = 4 * zoom * zoomRatio;
		group.scaling = zoom * zoomRatio;
		
		var bbox = new scope.Rectangle(group.bounds);
		var margin = 2;
		bbox.x = bbox.x - margin;
		bbox.y = bbox.y - margin;
		bbox.width = bbox.width + margin * 2;
		bbox.height = bbox.height + margin * 2;
		group.insertChild(0, new scope.Path.Rectangle(bbox));
		scope.view.viewSize = new scope.Size(bbox.size);
		group.position = scope.view.center;
	
		scope.view.draw();
		var raster = group.rasterize();
		var data = raster.getImageData(new scope.Rectangle(new scope.Point(0,0), raster.size));
	
		var edgeMagnitude = 0;
		for( var y = 1; y < data.height - 1; y++){
			var i = data.width * (y-1);
			for( var x = 1; x < data.width - 1; x++){
				// sobel filter
				var d00 = data.data[i*4];
				var d10 = data.data[(i+1)*4];
				var d20 = data.data[(i+2)*4];
				var d01 = data.data[(i+data.width)*4];
				var d21 = data.data[(i+data.width+2)*4];
				var d02 = data.data[(i+data.width*2)*4];
				var d12 = data.data[(i+data.width*2+1)*4];
				var d22 = data.data[(i+data.width*2+2)*4];
	
				var sobelx = -d00 + d20 - d01*2 + d21*2 - d02 + d22;
				var sobely = -d00 - d10*2 - d20 + d02 + d12*2 + d22;
				var sobel = Math.sqrt(sobelx*sobelx + sobely*sobely)/8;
				//data.data[(i+data.width+1)*4 + 1] = sobel;
				edgeMagnitude += sobel;
				i++;
			}
		}
		//for( var i = 0; i < data.width * data.height; i++){
		//	data.data[i*4] = 0;
		//	data.data[i*4+2] = 128;
		//	data.data[i*4+3] = 255;
		//}

		var coordDev = [];
		if( isCalcCoordDev ){
			// area cover ratio
			var sums = [];
			var w = Math.max(Math.floor(raster.width / 3), Math.floor(raster.height / 3));
			for( var iy = 0; iy < 3; iy++ ){
				var y0 = Math.max(Math.floor(raster.height / 2 + w * (iy - 1.5)), 0);
				var y1 = Math.min(Math.floor(raster.height / 2 + w * (iy - 0.5)), raster.height);
				for( var ix = 0; ix < 3; ix++ ){
					var x0 = Math.max(Math.floor(raster.width / 2 + w * (ix - 1.5)), 0);
					var x1 = Math.min(Math.floor(raster.width / 2 + w * (ix - 0.5)), raster.width);
					var sum = 0;
					for( var y = y0; y < y1; y++){
						var i = (y * raster.width + x0) * 4;
						for( var x = x0; x < x1; x++){
							sum += data.data[i];
							//data.data[i] = 25*(iy*3+ix);
							i += 4;
						}
					}
					sums.push(sum / (w * w * 255));
				}
			}
			coordDev.push((sums[0] + sums[1] + sums[2]) / 3);
			coordDev.push((sums[3] + sums[4] + sums[5]) / 3);
			coordDev.push((sums[6] + sums[7] + sums[8]) / 3);
			coordDev.push((sums[0] + sums[3] + sums[6]) / 3);
			coordDev.push((sums[1] + sums[4] + sums[7]) / 3);
			coordDev.push((sums[2] + sums[5] + sums[8]) / 3);
		}

		//raster.setImageData(data, new scope.Point(0,0));
		//var url = raster.toDataURL();

		raster.remove();
		group.remove();

		var edgeScore = (zoom * this._surfaceNormalize * normalize) / edgeMagnitude;
		return {'edgeScore': edgeScore, 'coordDev': coordDev,};// 'url': url};
	}

	buildFeatureGauge(features){
		// initialize gauge
		var gauge = [];
		for( var i = 0; i < this._nodeNum; i++ ){
			gauge.push({
				edgeScore: 1 - 0.5 * i / this._nodeNum,
				reducedEdgeScore: 1 - 0.5 * i / this._nodeNum,
				coordDev: [0.5, 0.5, 0.5, 0.5, 0.5, 0.5],
			})
		}

		// learning SOM
		var ratioViaIndex = [];
		var c = this._nodeNum / 10;
		for( i = 0; i < c; i++ ){
			ratioViaIndex.push(Math.exp(-i*i / (2*c*c/16))); // 4 sigma
		}

		for( var n = 0; n < this._epochNum; n++ ){
			var lr = 1 - (n / this._epochNum) * (1 - this._learningRate);
			var feature = features[Math.floor(Math.random() * features.length)];
			var minIndex = this._getMinIndex(feature, gauge).minIndex;
			var index0 = Math.max(minIndex - ratioViaIndex.length + 1, 0);
			var index1 = Math.min(minIndex + ratioViaIndex.length, gauge.length);
			for( var i = index0; i < index1; i++ ){
				this._fixGaugeFeature(feature, gauge[i], ratioViaIndex[Math.abs(i - minIndex)] * lr);
			}
		}

		return gauge;
	}

	applyFeatureGauge(silhouettes, features, gauge){
		// projection
		var indexes = [];
		for( var i = 0; i < features.length; i++ ){
			var minInfo = this._getMinIndex(features[i], gauge);
			var tmpIndex, nextIndex = -1;
			var tmpDistance, nextDistance;

			tmpIndex = minInfo.minIndex - 1;
			if( tmpIndex >= 0 ){
				tmpDistance = this._calcFeatureDistance(features[i], gauge[tmpIndex]);
				nextIndex = tmpIndex;
				nextDistance = tmpDistance;
			}
			tmpIndex = minInfo.minIndex + 1;
			if( tmpIndex < gauge.length ){
				tmpDistance = this._calcFeatureDistance(features[i], gauge[tmpIndex]);
				if( tmpDistance < nextIndex || nextIndex < 0 ){
					nextIndex = tmpIndex;
					nextDistance = tmpDistance;
				}
			}
			var denominator = minInfo.minDistance + nextDistance;
			silhouettes[i].location =
				minInfo.minIndex * nextDistance / denominator +
				nextIndex * minInfo.minDistance / denominator;
		}

		// sort
		silhouettes.sort(function(a, b){ return a.location - b.location; });
		silhouettes.forEach(function(silhouette){delete silhouette.location});
	}

	_getMinIndex(feature, gauge){
		var minIndex = -1;
		var minDistance = 8;
		for( var i = 0; i < gauge.length; i++ ){
			var distance = this._calcFeatureDistance(feature, gauge[i]);
			if( distance < minDistance ){
				minIndex = i;
				minDistance = distance;
			}
		}
		return {'minIndex': minIndex, 'minDistance': minDistance};
	}

	_calcFeatureDistance(refference, target){
		return(
			Math.pow(target.edgeScore - refference.edgeScore, 2)*2 +
			Math.pow(target.reducedEdgeScore - refference.reducedEdgeScore, 2)*2 +
			Math.pow(target.coordDev[0]- refference.coordDev[0], 2) +
			Math.pow(target.coordDev[1]- refference.coordDev[1], 2) +
			Math.pow(target.coordDev[2]- refference.coordDev[2], 2) +
			Math.pow(target.coordDev[3]- refference.coordDev[3], 2) +
			Math.pow(target.coordDev[4]- refference.coordDev[4], 2) +
			Math.pow(target.coordDev[5]- refference.coordDev[5], 2)
		);		
	}

	_fixGaugeFeature(refference, target, ratio){
		target.edgeScore = target.edgeScore * (1 - ratio) + refference.edgeScore * ratio;
		target.reducedEdgeScore = target.reducedEdgeScore * (1 - ratio) + refference.reducedEdgeScore * ratio;
		target.coordDev[0] = target.coordDev[0] * (1 - ratio) + refference.coordDev[0] * ratio;
		target.coordDev[1] = target.coordDev[1] * (1 - ratio) + refference.coordDev[1] * ratio;
		target.coordDev[2] = target.coordDev[2] * (1 - ratio) + refference.coordDev[2] * ratio;
		target.coordDev[3] = target.coordDev[3] * (1 - ratio) + refference.coordDev[3] * ratio;
		target.coordDev[4] = target.coordDev[4] * (1 - ratio) + refference.coordDev[4] * ratio;
		target.coordDev[5] = target.coordDev[5] * (1 - ratio) + refference.coordDev[5] * ratio;
	}
}
